#! /usr/bin/wish
#
# mastermind.tcl : マスターマインド
#
#                  Copyright (C) 1999,2000,2014,2019 by Makoto Hiroi
#
#
# global 変数
#      perm_table : 順列を格納するテーブル（リストで格納する）
#      select     : 選択した順列に印をつける
#      use_table  : 順列生成時に使用する
#      perm_count : 順列生成時に使用する
#      temp       : 順列生成時に使用する配列
#      ans_table  : 結果を格納するテーブル {code bulls cows} のリスト
#      ans_count  : 回数
#      level      : COM のレベル
#                   1 : ４回に３回はコードのチェックを行わない
#                   2 : ２回に１回はコードのチェックを行わない
#                   3 : コードのチェックを必ず行う
#      human_code : 人間側が当てるコード
#      com_code   : コンピュータ側が当てるコード
#      input_data : 入力データ管理用配列
#                   count  データ数
#                   data   データ（リスト）
#      play_flag  : ゲーム中の時は 1, そうでない場合は 0
#                   COM が当てたらビット１を立て、Human が当てたらビット２を立てる
#

# 順列の生成（本体）
proc make_perm_sub {n} {
    global perm_table perm_count use_table temp
    for {set i 0} {$i < 10} {incr i} {
	if {$use_table($i) == 0} {
	    set temp($n) $i
	    set m [expr $n + 1]
	    if {$m == 4} {
		# 作成終了
		set perm_table($perm_count) \
			[list $temp(0) $temp(1) $temp(2) $temp(3)]
		incr perm_count
	    } else {
		# 再帰するよ
	        set use_table($i) 1
	        make_perm_sub [expr $n + 1]
		set use_table($i) 0
	    }
	}
    }
}

# 順列の生成
proc make_perm {} {
    global perm_count use_table
    set perm_count 0
    for {set i 0} {$i < 10} {incr i} {
	set use_table($i) 0
    }
    make_perm_sub 0
}

# bulls を数える
proc count_bulls {code1 code2} {
    set bulls 0
    for {set i 0} {$i < 4} {incr i} {
	if {[lindex $code1 $i] == [lindex $code2 $i]} {
	    incr bulls
	}
    }
    return $bulls
}

# 同じ数字を数える
proc count_same_number {code1 code2} {
    set count 0
    foreach n $code1 {
	foreach m $code2 {
	    if {$n == $m} {
		incr count
		break
	    }
	}
    }
    return $count
}

# 選んだコードが矛盾していないか
proc check_code {code} {
    global ans_table ans_count
    for {set i 0} {$i < $ans_count} {incr i} {
	set old_code [lindex $ans_table($i) 0]
	set bulls [count_bulls $old_code $code]
	set cows  [expr [count_same_number $old_code $code] - $bulls]
	if { $bulls != [lindex $ans_table($i) 1] || \
	     $cows  != [lindex $ans_table($i) 2] } {
            return 0
        }
    }
    return 1
}

# コードを選ぶ
proc select_code {} {
    global perm_table select
    set n [expr int( rand() * 5040 )]
    for {set i $n} {$i >= 0} {incr i -1} {
	if {$select($i) == 0} {
	    set select($i) 1
	    return $perm_table($i)
	}
    }
    for {set i [expr $n + 1]} {$i < 5040} {incr i} {
	if {$select($i) == 0} {
	    set select($i) 1
	    return $perm_table($i)
	}
    }
    # ここにくるとエラーだよ
    return 0
}

# メッセージの表示
proc display_message {mes} {
    global input_data
    # ラベル .l0 に表示する
    set input_data(data) $mes
    update
    # 500 msec 待つ
    after 500
}

# COM の思考
proc think {} {
    global level ans_table ans_count com_code com_output play_flag
    display_message "私の番です"
    while 1 {
	set code [select_code]
        if {($level == 1 && [expr $ans_count % 4] != 0) || \
	    ($level == 2 && [expr $ans_count % 2] == 0) } {
            # チェックしない
	    break
        }
	if [check_code $code] break
    }
    # 問い合わせる
    set bulls [count_bulls $code $com_code]
    set cows  [expr [count_same_number $code $com_code] - $bulls]
    # 結果を格納する
    set ans_table($ans_count) [list $code $bulls $cows]
    # 表示する
    set com_output($ans_count) "  $code : bulls $bulls, cows $cows"
    if {$bulls == 4} {
	# ゲーム終了
	set play_flag [expr $play_flag | 0x02]
    }
    display_message "あなたの番です"
}

# ゲームの初期化
proc init_game {} {
    global ans_count select human_output com_output
    global input_data human_code com_code perm_table
    global space_mes
    set ans_count 0
    set input_data(count) 0
    set input_data(data) ""
    for {set i 0} {$i < 5040} {incr i} {
        set select($i) 0
    }
    for {set i 0} {$i < 10} {incr i} {
	set human_output($i) $space_mes
	set com_output($i) $space_mes
    }
    # コードの決定
    set n [expr int( rand() * 5040 )]
    set human_code $perm_table($n)
    set n [expr int( rand() * 5040 )]
    set com_code $perm_table($n)
}

# ゲームの開始
proc new_game {} {
    global play_flag
    if $play_flag {
	set ans [tk_messageBox -type yesno -icon question \
                   -message "今のゲームを中断しますか？"]
        if {$ans == "no"} return
    }
    # ゲーム開始
    set play_flag 1
    init_game
    # COM の思考
    think
}

# ********** ボタンの処理 **********

# 数字を押した場合
proc push_number {n} {
    global input_data play_flag
    if $play_flag {
        if {$input_data(count) == 0} {
	    set input_data(data) ""
        }
        if { ($input_data(count) < 4) && $play_flag } {
	    foreach m $input_data(data) {
	        if { $m == $n } return
	    }
	    set input_data(data) [concat $input_data(data) $n]
	    incr input_data(count)
        }
    }
}

# CLR を押した場合
proc push_clr {} {
    global input_data play_flag
    if $play_flag {
      set input_data(count) 0
      set input_data(data) ""
    }
}

# ゲーム終了
proc game_over {} {
    global play_flag human_code
    if {$play_flag == 5} {
	set mes "あなたの勝ちです"
    } elseif {$play_flag == 3} {
	set mes "私の勝ちです、正解は $human_code でした"
    } else {
	set mes "引き分けです"
    }
    display_message $mes
    set play_flag 0
}

# OK を押した場合
proc push_ok {} {
    global input_data human_code human_output ans_count play_flag
    if { $play_flag && $input_data(count) == 4 } {
	set bulls [count_bulls $input_data(data) $human_code]
	set cows  [expr [count_same_number $input_data(data) $human_code] - $bulls]
        set human_output($ans_count) "  $input_data(data) : bulls $bulls, cows $cows"
	if {$bulls == 4 } {
	    set play_flag [expr $play_flag | 0x04]
	}
	if {$play_flag > 1} {
	    game_over
	    return
	} else {
	    set input_data(data) ""
	    set input_data(count) 0
	}
	incr ans_count
	if {$ans_count >= 10} {
	    game_over
	}
	# COM の思考
	think
    }
}

# ヘルプファイルの表示
proc help {} {
    global argv0
    if {![winfo exist .t0]} {
        toplevel .t0
        wm title .t0 "MasterMind Help"
        text .t0.text -yscrollcommand ".t0.scroll set"
        scrollbar .t0.scroll -command ".t0.text yview"
	pack .t0.scroll -side right -fill y
	pack .t0.text -side left
	# 起動パスを取得
	set path [file dirname $argv0]
	# ファイルの読み込み
	set f [open "$path/mastermind.txt" r]
	while {![eof $f]} {
	    .t0.text insert end [read $f 1000]
	}
	close $f
    }
}


# ********** 画面の生成 **********
wm geometry . 460x420

# ********** メニューの設定 **********

menu .m -type menubar
. configure -menu .m -bg darkgreen
.m add cascade -label "Games" -under 0 -menu .m.m1
.m add cascade -label "Level" -under 0 -menu .m.m2
.m add command -label "Help" -under 0 -command "help"
menu .m.m1 -tearoff no
.m.m1 add command -label "Start" -under 0 -command "new_game"
.m.m1 add command -label "Exit" -under 0 -command "exit"
menu .m.m2 -tearoff no
.m.m2 add radiobutton -label "Level 1" -variable level -value 1
.m.m2 add radiobutton -label "Level 2" -variable level -value 2
.m.m2 add radiobutton -label "Level 3" -variable level -value 3

# フォントの変更
option add *font "{Noto Sans Mono CJK JP} 12"
# ラベルの生成
label .l0 -font "{Noto Sans Mono CJK JP} 16" -textvariable input_data(data) -bg darkgreen -fg white
# ボタンの生成
frame .f0
button .b1  -text "1" -command "push_number 1"
button .b2  -text "2" -command "push_number 2"
button .b3  -text "3" -command "push_number 3"
button .b4  -text "4" -command "push_number 4"
button .b5  -text "5" -command "push_number 5"
button .b6  -text "6" -command "push_number 6"
button .b7  -text "7" -command "push_number 7"
button .b8  -text "8" -command "push_number 8"
button .b9  -text "9" -command "push_number 9"
button .b10 -text "0" -command "push_number 0"
button .b11 -text "OK" -command "push_ok"
button .b12 -text "AC" -command "push_clr"

for {set i 1} {$i <= 12} {incr i} {
    pack .b$i -in .f0 -side left
}

# 出力用ラベルの生成
frame .f1
frame .f2
label .l1 -text "    コンピュータの結果     " -fg yellow -bg darkgreen
pack .l1 -in .f1
label .l2 -text "       あなたの結果        " -fg white  -bg darkgreen
pack .l2 -in .f2
set space_mes   "                           "
for {set i 0} {$i < 10} {incr i 1} {
    label .lc$i -textvariable com_output($i) -bg darkgreen -fg yellow
    set com_output($i) $space_mes
    pack .lc$i -in .f1
    label .lh$i -textvariable human_output($i) -bg darkgreen -fg white
    set human_output($i) $space_mes
    pack .lh$i -in .f2
}

pack .l0 .f0
pack .f1 .f2 -side left

# グローバル変数の初期化
set perm_count 0
set play_flag 0
set level 1
# 順列の生成
make_perm

# ********** end of file **********
