#! /usr/bin/wish
#
# spon.tcl : そろえてポン
#
#            Copyright (C) 1998 - 2000, 2014, 2019 by Makoto Hiroi
#
# global変数
#    board 盤面（配列）   グラフィック部品の id を格納
#         0  1  2  3  4
#         5  6  7  8  9
#        10 11 12 13 14
#        15 16 17 18 19 
#        20 21 22 23 24
#
#    value        (配列) 盤面の色を数値で格納
#    color_table  (配列) 色と数値の対応
#    change_table (配列) クリックした時に色が変わる位置
#    color_number        色の種類
#    clear_data  （配列）解けた問題を格納 (文字列データ: レベル 色数 配置)
#    min_data    （配列）最小ステップ数を格納
#    level               3 : Easy, 2 : Normal, 1 : Hard
#    start_data          初期状態を格納
#    min_step            最小ステップ数
#    now_step            現在のステップ数
#    output_label        ステップ数表示用変数
#    select_num          選択した問題(-1:new game)
#

# **********押した位置に対する変更する位置 **********
set change_table(0) {0 1 5}
set change_table(1) {0 1 2 6}
set change_table(2) {1 2 3 7}
set change_table(3) {2 3 4 8}
set change_table(4) {3 4 9}
set change_table(5) {0 5 6 10}
set change_table(6) {1 5 6 7 11}
set change_table(7) {2 6 7 8 12}
set change_table(8) {3 7 8 9 13}
set change_table(9) {4 8 9 14}
set change_table(10) {5 10 11 15}
set change_table(11) {6 10 11 12 16}
set change_table(12) {7 11 12 13 17}
set change_table(13) {8 12 13 14 18}
set change_table(14) {9 13 14 19}
set change_table(15) {10 15 16 20}
set change_table(16) {11 15 16 17 21}
set change_table(17) {12 16 17 18 22}
set change_table(18) {13 17 18 19 23}
set change_table(19) {14 18 19 24}
set change_table(20) {15 20 21}
set change_table(21) {16 20 21 22}
set change_table(22) {17 21 22 23}
set change_table(23) {18 22 23 24}
set change_table(24) {19 23 24}

# ********** 色のテーブル **********
set color_table(0) white
set color_table(1) yellow
set color_table(2) green
set color_table(3) blue
set color_table(4) red

# global の初期化
proc init_global {} {
    global color_number rect_size level select_num
    set level 3
    set color_number 2
    set rect_size 50
    set select_num -1
}

# ヘルプファイルの表示
proc help {} {
    global argv0
    if {![winfo exist .t0]} {
        toplevel .t0
        wm title .t0 "Help File"
        text .t0.text -yscrollcommand ".t0.scroll set" -font "{Noto Sans Mono CJK JP} 12"
        scrollbar .t0.scroll -command ".t0.text yview"
	pack .t0.scroll -side right -fill y
	pack .t0.text -side left
	# 起動パスを取得
	set path [file dirname $argv0]
	# ファイルの読み込み
	set f [open "$path/spon.txt" r]
	while {![eof $f]} {
	    .t0.text insert end [read $f 1000]
	}
	close $f
    }
}

# クリアデータを読み込む
proc read_data {} {
    global argv0 clear_data min_data
    # spon.tcl の起動パスを取得する
    set path [file dirname $argv0]
    if [file exist "$path/spon.dat"] {
	# spon.data をリードオープン
	set f [open "$path/spon.dat" r]
	set i 0
	while {[gets $f line] >= 0} {
	    set data [split $line " "]
	    set clear_data($i) [lindex $data 0]
	    set min_data($i) [lindex $data 1]
	    incr i
	}
	close $f
    }
}

# クリアデータを書き込む
proc write_data {} {
    global argv0 start_data clear_data
    global min_step min_data select_num
    # spon.tcl の起動パスを取得する
    set path [file dirname $argv0]
    set j [array size clear_data]
    if {$select_num >= 0} {
	if {$min_step < $min_data($select_num)} {
	    set min_data($select_num) $min_step
	    # データをライトする
	    set f [open "$path/spon.dat" w]
	    for {set i 0} {$i < $j} {incr i} {
		puts $f "$clear_data($i) $min_data($i)"
	    }
	    close $f
	}
    } else {
	# クリアデータをセット
	set clear_data($j) $start_data
	set min_data($j)   $min_step
	# spon.data の最後にデータを追加
	set f [open "$path/spon.dat" a]
	puts $f "$start_data $min_step"
	close $f
    }
}

# 盤面を作る
proc make_board {} {
    global color_number level
    for {set i 0} {$i < 25} {incr i} {
	set use_table($i) 0
    }
    for {set i 0} {$i < 25} {incr i $level} {
	while 1 {
	    # 押していない場所を選ぶ
	    set p [expr int(rand() * 25)]
	    if {$use_table($p) == 0} {
		set use_table($p) 1
		break
	    }
	}
	# 押す回数を選ぶ
	set n [expr int(rand() * $color_number)]
	while {$n > 0} {
	    change_color $p
	    incr n -1
	}
    }
}

# パネルの色をセット
proc set_color {n c} {
    global value board color_table
    .c0 itemconfigure $board($n) -fill $color_table($c)
    set value($n) $c
}

# 色の変更
proc change_color {postion} {
    global color_number change_table value
    foreach n $change_table($postion) {
	set_color $n [expr ($value($n) + 1) % $color_number]
    }
}

# 全部解けた？
proc check_finish {} {
    global value
    set c $value(0)
    for {set x 1} {$x < 25} {incr x} {
	if {$c != $value($x)} {
	    return 0
	}
    }
    return 1
}


# 色を押したか
proc push_rectangle {x0 y0} {
    global rect_size min_step now_step clear_data
    set x [expr $x0 / $rect_size]
    set y [expr $y0 / $rect_size]
    change_color [expr $y * 5 + $x]
    incr now_step
    display_step
    if [check_finish] {
        if {$min_step > $now_step} {
	    set min_step $now_step
	}
	# クリアデータセーブ
        write_data
	tk_messageBox -type ok -message "おめでとう"
	retry_game
    }
}

# 初期状態を記憶
proc save_start {} {
    global value level color_number start_data
    switch $level {
	1 {set result "H"}
	2 {set result "N"}
	3 {set result "E"}
    }
    set result [concat $result $color_number]
    for {set i 0} {$i < 25} {incr i} {
	set result [concat $result $value($i)]
    }
    # 文字列に変換して格納する
    set start_data [join $result ""]
}

# 問題のチェック    
proc check_clear_data {} {
    global start_data clear_data
    # 配列にデータがない（存在しない）場合は 0 が返る
    set size [array size clear_data]
    for {set i 0} {$i < $size} {incr i} {
	if {![string compare $start_data $clear_data($i)]} {
	    return 0
	}
    }
    return 1
}

# 選択したゲームをセットする
proc set_game {n} {
    global clear_data min_data min_step now_step
    global level color_number select_num start_data
    set select_num $n
    set start_data $clear_data($n)
    set min_step $min_data($n)
    set now_step 0
    set data [split $clear_data($n) ""]
    switch [lindex $data 0] {
	H {set level 1}
	N {set level 2}
	E {set level 3}
    }
    set color_number [lindex $data 1]
    for {set i 0; set j 2} {$i < 25} {incr i; incr j} {
	set_color $i [lindex $data $j]
    }
    display_step
}

# ゲームの選択
proc select_game {} {
    global board value clear_data min_data
    toplevel  .sw
    wm title  .sw "GameSelect"
    listbox   .sw.lb -yscrollcommand ".sw.scroll set" -width 32
    scrollbar .sw.scroll -command ".sw.lb yview"
    pack .sw.scroll -side right -fill y
    pack .sw.lb     -side left
    for {set i 0} {$i < [array size clear_data]} {incr i} {
	.sw.lb insert end "$clear_data($i) $min_data($i)"
    }
    bind .sw.lb <Double-1> {
	set_game [.sw.lb index active]
    }
}

# 最初からやり直し
proc retry_game {} {
    global start_data now_step
    set now_step 0
    set data [split $start_data ""]
    for {set i 0; set j 2} {$i < 25} {incr i; incr j} {
	set_color $i [lindex $data $j]
    }
    display_step
}


# 新しいゲームへ
proc new_game {} {
    global board value now_step min_step select_num
    # 白に戻す
    for {set i 0} {$i < 25} {incr i} {
	.c0 itemconfigure $board($i) -fill white
        set value($i) 0
    }
    set select_num -1
    set now_step 0
    set min_step 999
    display_step
    while 1 {
        # 盤面を作る
        make_board
	# 初期状態を記憶
	save_start
	if {![check_finish] && [check_clear_data]} {
	    break
	}
    }
}    

# ステップ数の表示
proc display_step {} {
    global min_step now_step output_label
    set output_label [format "Min Steps %3d : Step %3d" $min_step $now_step]
}

# ********** 画面の設定 **********
canvas .c0 -width 250 -height 250
wm title . "SoroetePon"
# ステップ表示用ラベルの設定
label .l0 -textvariable output_label -font "{Noto Sans Mono CJK JP} 12"
pack .l0 .c0

# ********** rectangle の作成 **********
set i 0
for {set y 0} {$y < 5} {incr y 1} {
    set y0 [expr $y * 50]
    set y1 [expr $y0 + 50]
    for {set x 0} {$x < 5} {incr x 1} {
        set x0 [expr $x * 50]
	set x1 [expr $x0 + 50]
	set board($i) [.c0 create rectangle $x0 $y0 $x1 $y1 -fill white]
        set value($i) 0
	incr i 1
    }
}

# メニューの設定
menu .m -type menubar
. configure -menu .m
.m add cascade -label "Games"  -under 0 -menu .m.m1
.m add cascade -label "Level"  -under 0 -menu .m.m2
.m add command -label "Retry"  -under 0 -command "retry_game" 
.m add command -label "Select" -under 0 -command "select_game"
.m add command -label "Help"   -under 0 -command "help"
menu .m.m1 -tearoff no
.m.m1 add command -label "NewGame" -under 0 -command "new_game"
.m.m1 add separator
.m.m1 add radiobutton -label "2 colors" -variable color_number -value 2 \
                      -under 0 -command "new_game"
.m.m1 add radiobutton -label "3 colors" -variable color_number -value 3 \
                      -under 0 -command "new_game"
.m.m1 add radiobutton -label "4 colors" -variable color_number -value 4 \
                      -under 0 -command "new_game"
.m.m1 add radiobutton -label "5 colors" -variable color_number -value 5 \
                      -under 0 -command "new_game"
.m.m1 add separator
.m.m1 add command -label "Exit" -under 0 -command "exit"
menu .m.m2 -tearoff no
.m.m2 add radiobutton -label "Easy"   -under 0 -variable level -value 3
.m.m2 add radiobutton -label "Normal" -under 0 -variable level -value 2
.m.m2 add radiobutton -label "Hard"   -under 0 -variable level -value 1

# ********** バインドの設定 **********
bind .c0  <Button-1>  "push_rectangle %x %y"

# **********グローバル変数の初期化 **********
init_global
new_game
read_data
# ********** end of file **********
