#! /usr/bin/wish
#
# TEN_SHAN.TCL : パズルゲーム「テン」真・上海モード
#
#                Copyright (C) 1999,2000,2014,2019 by Makoto Hiroi
#

# 共通ライブラリロード
source "ten_sub.tcl"

# 盤面から色を求める
proc get_color {id} {
    global board
    return [expr $board($id) / 16]
}

# 盤面から数字を求める
proc get_number {id} {
    global board
    return [expr $board($id) % 16]
}

# 選択できるか
proc check_select {id} {
    global board neighbor_up neighbor_left neighbor_right
    if {$board($id) != 0 && $board($neighbor_up($id)) == 0} {
	if {$board($neighbor_left($id)) == 0 || $board($neighbor_right($id)) == 0} {
	    return 1
	}
    }
    return 0
}

# 取れる牌を探す
proc search_piece {} {
    global board piece piece_postion piece_count
    set result ""
    for {set c 0} {$c < 4} {incr c} {
	for {set n 1} {$n < 5} {incr n} {
	    set p1 [expr $c * 16 + $n]
	    set p2 [expr $c * 16 + (10 - $n)]
	    if {$piece_count($p1) > 0 && $piece_count($p2) > 0} {
		foreach id1 $piece_postion($p1) {
		    if ![check_select $id1] continue
		    foreach id2 $piece_postion($p2) {
			if [check_select $id2] {
			    set result [concat $result $piece($id1) $piece($id2)]
			}
		    }
		}
	    }
	}
	# 5 だけ特別
	set p1 [expr $c * 16 + 5]
	if {$piece_count($p1) > 0} {
	    set i 0
	    while {$i < 4} {
		set id1 [lindex $piece_postion($p1) $i]
		incr i
		if ![check_select $id1] continue
		set j $i
		while {$j < 4} {
		    set id2 [lindex $piece_postion($p1) $j]
		    incr j
		    if [check_select $id2] {
			set result [concat $result $piece($id1) $piece($id2)]
		    }
		}
	    }
	}
    }
    return $result
}

# 終了チェック
proc check_finish {} {
    global board rest time id buff2 play_flag
    if {$rest == 0} {
	# 終了
	set t [clock seconds]
	set s [expr $t - $time]
	after cancel $id
	set ranking [check_hi_score $s]
	if {$ranking > 0 && $play_flag == 1} {
	    input_hi_score_name $ranking 
	    tkwait window .t2
	    update_score $buff2 $t $s $ranking
	    write_score_file
	    open_score_window $ranking
	} else {
	    tk_messageBox -type ok \
		-message [format "おめでとう %s です" [change_seconds $s]]
	}
	return 1
    } else {
	# 手詰まりチェック
	if {[search_piece] == ""} {
	    after cancel $id
	    tk_messageBox -type ok -message "手詰まりです"
	    return 1
	}
    }
    return 0
}

# 牌を削除する
proc remove_piece {id} {
    global board piece number rest piece_count history
    .c0 lower $number($id)
    .c0 lower $piece($id)
    .c0 itemconfigure $piece($id) -fill white
    set p $board($id)
    incr piece_count($p) -1
    set history($rest) [concat $p $id]
    set board($id) 0
    incr rest -1
    if {$id == 127 || $id == 131} {
	if {$board(127) == 0 && $board(131) == 0} {
	    set board(127_131) 0
	}
    }
    if {$id == 130 || $id == 134} {
	if {$board(130) == 0 && $board(134) == 0} {
	    set board(130_134) 0
	}
    }
}


# バインド
proc push_piece {id} {
    global board piece number play_flag pid
    if {!($play_flag & 0x03) || ![check_select $id]} return
    if {$pid == -1} {
	# 最初の選択
	set pid $id
	.c0 itemconfigure $piece($id) -fill darkgray
    } elseif {$pid == $id} {
	# 同じであればキャンセル扱いにする
	.c0 itemconfigure $piece($id) -fill white
	set pid -1
    } else {
	set n1 [get_number $pid]
	set c1 [get_color $pid]
	set n2 [get_number $id]
	set c2 [get_color $id]
	if {$c1 == $c2 && [expr $n1 + $n2] == 10} {
	    .c0 itemconfigure $piece($id) -fill darkgray
	    update
	    after 200
	    # 削除する
	    remove_piece $id
	    remove_piece $pid
	    set pid -1
	    display_message
	}
	# 手詰まりのチェック
	if [check_finish] {
	    set play_flag 0
	}
    }
}

# 牌を戻す
proc takeback {} {
    global play_flag number piece board rest history pid back
    if {$play_flag != 0 && $pid != -1} {
	# 牌を選択している時はキャンセル扱いにする
	push_piece $pid
	return
    }
    if {!($play_flag & 0x03) || $rest == 144} return
    set play_flag 2
    for {set i 0} {$i < 2} {incr i} {
	incr rest
	set h $history($rest)
	set p  [lindex $h 0]
	set id [lindex $h 1]
	set id1 [expr $id - 1]
	while {$id1 >= 0} {
	    if {$board($id1) != 0} break
	    incr id1 -1
	}
	if {$id1 > 0} {
	    .c0 raise $piece($id) $number($id1)
	    .c0 raise $number($id) $piece($id)
	} else {
	    .c0 raise $piece($id) $back
	    .c0 raise $number($id) $piece($id)
	}
	set board($id) $p
	if {$id == 127 || $id == 131} {
	    set board(127_131) 1
	} elseif {$id == 130 || $id == 134} {
	    set board(130_134) 1
	}
    }
}

# 検索
proc search {} {
    global play_flag pid
    if {!($play_flag & 0x03)} return
    if {$pid != -1} {
	# 選択していたらキャンセルして表示する
	push_piece $pid
    }
    set play_flag 4
    set pieces [search_piece]
    set len [llength $pieces]
    set i 0
    while {$i < $len} {
	set p1 [lindex $pieces $i]
	incr i
	set p2 [lindex $pieces $i]
	incr i
	.c0 itemconfigur $p1 -fill darkgray
	.c0 itemconfigur $p2 -fill darkgray
	update
	after 1000
	.c0 itemconfigur $p1 -fill white
	.c0 itemconfigur $p2 -fill white
    }
    set play_flag 2
}

# ********** 初期化 **********

# 盤面の初期化
proc make_board {} {
    global board piece_postion piece_count
    set board(-1) 0
    set board(130_134) 1
    set board(127_131) 1
    for {set i 0; set c 0} {$c < 4} {incr c} {
	for {set n 1} {$n <= 9} {incr n} {
	    # piece_postion の初期化
	    set p [expr $c * 16 + $n]
	    set piece_postion($p) ""
	    set piece_count($p) 4
	    for {set j 0} {$j < 4} {incr j} {
		set board($i) $p
		incr i
	    }
	}
    }
    # 乱数でかき回す
    for {set j 0} {$j < $i} {incr j} {
	set n    [expr int( rand() * $i )]
	set temp $board($n)
	set board($n) $board($j)
	set board($j) $temp
    }
    # piece_postion のセット
    for {set i 0} {$i < 144} {incr i} {
	set p $board($i)
	set piece_postion($p) [concat $piece_postion($p) $i]
    }
}

# 盤面の描画
proc draw_board {} {
    global number numstr piece color
    for {set i 0} {$i < 144} {incr i} {
	set n [get_number $i]
	set c [get_color $i]
	.c0 itemconfigure $number($i) -text $numstr($n) -fill $color($c)
	.c0 raise $piece($i)
	.c0 raise $number($i)
    }
}

# 周辺部のデータ
set another_data(1) {2 3 12 13}
set another_data(2) {}
set another_data(3) {3 12}
set another_data(4) {2 3 12 13}
set another_data(5) {2 3 12 13}
set another_data(6) {3 12}
set another_data(7) {}
set another_data(8) {2 3 12 13}

# ピラミッド以外の作成
proc make_another {} {
    global number piece another_data 
    global neighbor_left neighbor_right neighbor_up
    # 頂上
    set x1 [expr 32 * 7 + 4]
    set x2 [expr $x1 + 31]
    set y1 [expr 32 * 4 + 4]
    set y2 [expr $y1 + 31]
    set piece(120) [.c0 create rectangle $x1 $y1 $x2 $y2 -fill white]
    set number(120) [.c0 create text [expr $x1 + 16] [expr $y1 + 16] \
	    -text " " -font "{Noto Sans Mono CJK JP} 24 bold"]
    .c0 bind $number(120) <Button-1> "push_piece 120"
    .c0 bind $piece(120) <Button-1> "push_piece 120"
    set neighbor_right(120) -1
    set neighbor_left(120) -1
    # 周辺部
    set id 121
    for {set y 1} {$y <= 8} {incr y} {
	set y1 [expr 32 * $y]
	set y2 [expr $y1 + 31]
	foreach x $another_data($y) {
	    set x1 [expr $x * 32]
	    set x2 [expr $x1 + 31]
	    set piece($id) [.c0 create rectangle $x1 $y1 $x2 $y2 -fill white]
	    set number($id) [.c0 create text [expr $x1 + 16] [expr $y1 + 16] \
		    -text " " -font "{Noto Sans Mono CJK JP} 24 bold"]
	    .c0 bind $number($id) <Button-1> "push_piece $id"
	    .c0 bind $piece($id) <Button-1> "push_piece $id"
	    incr id
	}
    }
    # 残り３枚
    set y1 [expr 32 * 4 + 16]
    set y2 [expr $y1 + 31]
    foreach x {1 14 15} {
	set x1 [expr 32 * $x]
	set x2 [expr $x1 + 31]
	set piece($id) [.c0 create rectangle $x1 $y1 $x2 $y2 -fill white]
	set number($id) [.c0 create text [expr $x1 + 16] [expr $y1 + 16] \
		-text " " -font "{Noto Sans Mono CJK JP} 24 bold"]
	.c0 bind $number($id) <Button-1> "push_piece $id"
	.c0 bind $piece($id) <Button-1> "push_piece $id"
	incr id
    }
    # 隣接データ
    set neighbor_left(121)  -1
    set neighbor_right(121) 122
    set neighbor_left(122)  121
    set neighbor_right(122) 0
    set neighbor_left(123)  7
    set neighbor_right(123) 124
    set neighbor_left(124)  123
    set neighbor_right(124) -1
    set neighbor_left(125)  -1
    set neighbor_right(125) 16
    set neighbor_left(126)  23
    set neighbor_right(126) -1
    set neighbor_left(127)  141
    set neighbor_right(127) 128
    set neighbor_left(128)  127
    set neighbor_right(128) 24
    set neighbor_left(129)  31
    set neighbor_right(129) 130
    set neighbor_left(130)  129
    set neighbor_right(130) 142
    set neighbor_left(131)  141
    set neighbor_right(131) 132
    set neighbor_left(132)  131
    set neighbor_right(132) 32
    set neighbor_left(133)  39
    set neighbor_right(133) 134
    set neighbor_left(134)  133
    set neighbor_right(134) 142
    set neighbor_left(135)  -1
    set neighbor_right(135) 40
    set neighbor_left(136)  47
    set neighbor_right(136) -1
    set neighbor_left(137)  -1
    set neighbor_right(137) 138
    set neighbor_left(138)  137
    set neighbor_right(138) 56
    set neighbor_left(139)  63
    set neighbor_right(139) 140
    set neighbor_left(140)  139
    set neighbor_right(140) -1
    set neighbor_left(141)  -1
    set neighbor_right(141) "127_131"
    set neighbor_left(142)  143
    set neighbor_right(142) "130_134"
    set neighbor_left(143)  142
    set neighbor_right(143) -1
    # 補正
    set neighbor_left(0) 122
    set neighbor_right(7) 123
    set neighbor_left(16) 125
    set neighbor_right(23) 126
    set neighbor_left(24) 128
    set neighbor_right(31) 129
    set neighbor_left(32) 132
    set neighbor_right(39) 133
    set neighbor_left(40) 135
    set neighbor_right(47) 136
    set neighbor_left(56) 138
    set neighbor_right(63) 139
    # 補正その２
    set neighbor_up(116) 120
    set neighbor_up(117) 120
    set neighbor_up(118) 120
    set neighbor_up(119) 120
}

# ピラミッドの１段を作る
proc make_floor {id w h} {
    global piece number 
    global neighbor_left neighbor_right neighbor_up
    set x0 [expr 32 * 4 + $h * 28]
    set y0 [expr 32 + $h * 28]
    set w0 [expr $w - 1]
    set up_id [expr $id + $w * $w]
    for {set y 0} {$y < $w} {incr y} {
	set y1 [expr $y0 + $y * 32]
	set y2 [expr $y1 + 31]
	for {set x 0} {$x < $w} {incr x} {
	    set x1 [expr $x0 + $x * 32]
	    set x2 [expr $x1 + 31]
	    set piece($id) [.c0 create rectangle $x1 $y1 $x2 $y2 -fill white]
	    set number($id) [.c0 create text [expr $x1 + 16] [expr $y1 + 16] \
		    -text " " -font "{Noto Sans Mono CJK JP} 24 bold"]
	    .c0 bind $number($id) <Button-1> "push_piece $id"
	    .c0 bind $piece($id) <Button-1> "push_piece $id"
	    # 隣接データセット
	    if {$x == 0} {
		set neighbor_right($id) [expr $id + 1]
		set neighbor_left($id) -1
	    } elseif {$x == $w0} {
		set neighbor_right($id) -1
		set neighbor_left($id) [expr $id - 1]
	    } else {
		set neighbor_right($id) [expr $id + 1]
		set neighbor_left($id)  [expr $id - 1]
	    }
	    # 上のカードをセット
	    if {$x > 0 && $x < $w0 && $y > 0 && $y < $w0} {
		set neighbor_up($id) $up_id
		incr up_id
	    }
	    incr id
	}
    }
}


# メッセージの表示
proc display_message {} {
    global time buff1 rest
    set t [expr [clock seconds] - $time]
    set buff1 [format "残り %3d 枚    時間 %5s" $rest [change_seconds $t]]
}

# 時間の表示
proc display_time {} {
    global id
    display_message
    set id [after 1000 display_time]
}

# ゲームの開始
proc start_game {} {
    global play_flag time rest id pid
    if {$play_flag > 0} {
	after cancel $id
    }
    while 1 {
	make_board
	if {[search_piece] != ""} break
    }
    draw_board
    set rest 144
    set play_flag 1
    set time [clock seconds]
    set pid -1
    display_time
}

# ********** メニューの設定 **********
menu .m -type menubar
. configure -menu .m
.m add cascade -label "Games"    -under 0 -menu .m.m1
.m add command -label "Takeback" -under 0 -command "takeback"
.m add command -label "Search"   -under 0 -command "search"
.m add command -label "Help"     -under 0 -command "help"
menu .m.m1 -tearoff no
.m.m1 add command -label "Start"   -under 0 -command "start_game"
.m.m1 add command -label "HiScore" -under 0 -command "open_score_window 0"
.m.m1 add separator
.m.m1 add command -label "Exit" -under 0 -command "exit"

# オプションの設定
option add *font "{Noto Sans Mono CJK JP} 12"

# 画面の設定
canvas .c0 -width 544 -height 320
set back [.c0 create rectangle 0 0 544 320 -fill darkgreen]

# ***** ピラミッドの部分を作る *****
# 初期化
for {set i 0} {$i < 144} {incr i} {
    set neighbor_up($i) -1
}
make_floor  0  8 0
make_floor 64  6 1
make_floor 100 4 2
make_floor 116 2 3

# その他の部分を作る
make_another

# 表示用ラベル
label .l1 -textvariable buff1 -bg darkgreen -fg white -anchor e

pack .l1 .c0 -fill x

# 窓の題名
wm title . "TEN"
wm resizable . 0 0

# 初期化
expr srand( [clock seconds] )
set play_flag 0
set path_name [file dirname $argv0]
set score_file "$path_name/TEN2上海.SCO"

# スコアファイルのリード
read_score_file

focus -force .

# end of file

