#
# TEN_SUB.TCL : パズルゲーム「ＴＥＮ」共通サブルーチン
#
#               Copyright (C) 1999,2000,2014 by Makoto Hiroi
#

# 色
set color(0) red
set color(1) blue
set color(2) green
set color(3) gold

# 数字
set numstr(0) ""
set numstr(1) "１"
set numstr(2) "２"
set numstr(3) "３"
set numstr(4) "４"
set numstr(5) "５"
set numstr(6) "６"
set numstr(7) "７"
set numstr(8) "８"
set numstr(9) "９"

# ヘルプファイルの表示
proc help {} {
    global path_name
    if {![winfo exist .t0]} {
        toplevel .t0
        wm title .t0 "Puzzle TEN Help"
        text .t0.text -yscrollcommand ".t0.scroll set"
        scrollbar .t0.scroll -command ".t0.text yview"
	pack .t0.scroll -side right -fill y
	pack .t0.text -side left
	# ファイルの読み込み
	set f [open "$path_name/ten_help.txt" r]
	while {![eof $f]} {
	    .t0.text insert end [read $f 1000]
	}
	close $f
	.t0.text configure -state disabled
    }
}

# スコアファイルリード
proc read_score_file {} {
    global score_file name date score
    set i 1
    if [file exists $score_file] {
	# ファイルの読み込み
	set f [open $score_file r]
	while {[gets $f line] >= 0} {
	    set l [split $line "\t"]
	    set name($i)  [lindex $l 0]
	    set date($i)  [lindex $l 1]
	    set score($i) [lindex $l 2]
	    incr i
	}
	close $f
    }
    set now_date [clock seconds]
    while {$i <= 10} {
	set name($i) ""
	set date($i) $now_date
	# 5999 は 99:59 です
	set score($i) 5999
	incr i
    }
}

# スコアファイルライト
proc write_score_file {} {
    global score_file name date score
    set f [open $score_file w]
    for {set i 1} {$i <= 10} {incr i} {
	puts $f [format "%s\t%d\t%d" $name($i) $date($i) $score($i)]
    }
    close $f
}


# ********** スコア表示 ********

# 秒数をスコアに変換
proc change_seconds {s} {
    return [format "%02d:%02d" [expr $s / 60] [expr $s % 60]]
}

#
# トップテンウィンドウを開く
# [注意]
# ver 8.1α ではユニコードを採用しているので、format の %s では
# 全角文字を 2 byte とカウントしない。このため、一つのラベルで表
# 示すると、全角文字で名前を入力されると表示位置がずれる場合がある。
# これを回避するため、順位と名前、スコアと日付を別ラベルで表示する。
#
proc open_score_window {ranking} {
    global name date score
    if [winfo exists .t1] {
	destroy .t1
    }
    toplevel .t1
    wm title .t1 "Top 10"
    frame .t1.f0
    frame .t1.f1
    label .t1.f0.l0 -text "順位  名前" -anchor w
    label .t1.f1.l0 -text "記録    日付  " -anchor w
    pack .t1.f0.l0 -fill x
    pack .t1.f1.l0 -fill x
    for {set i 1} {$i <= 10} {incr i} {
	label .t1.f0.l$i -text [format "%4d %-20s" $i $name($i)] -anchor w
	label .t1.f1.l$i -text [format "%5s %8s" \
                [change_seconds $score($i)] \
                [clock format $date($i) -format "%y/%m/%d"]]
	pack  .t1.f0.l$i -fill x
	pack  .t1.f1.l$i -fill x
    }
    if {$ranking > 0} {
	.t1.f0.l$ranking configure -fg red
	.t1.f1.l$ranking configure -fg red
    }
    pack .t1.f0 .t1.f1 -side left
}

# ベストテンに入るか
proc check_hi_score {now_score} {
    global score
    for {set i 1} {$i <= 10} {incr i} {
	if {$score($i) > $now_score} {
	    # ベストテンに入ったよ
	    return $i
	}
    }
    return 0
}

# スコアの更新
proc update_score {n d s o} {
    global name date score
    for {set i 9} {$i >= $o} {incr i -1} {
	set j [expr $i + 1]
	set name($j) $name($i)
	set date($j) $date($i)
	set score($j) $score($i)
    }
    set name($o) $n
    set date($o) $d
    set score($o) $s
}

# トップテンの名前入力
proc input_hi_score_name {ranking} {
    global buff2
    set buff2 ""
    toplevel .t2
    wm title .t2 "Input Your Name"
    wm geometry .t2 "+[expr [winfo x .] + 120]+[expr [winfo y .] + 180]"
    label .t2.l0 -text [format "おめでとう！ %d 位です" $ranking]
    label .t2.l1 -text "名前を入力してね"
    entry .t2.e0 -textvariable buff2
    focus -force .t2.e0
    # grab set -global .t2
    bind .t2.e0 <Return> {
	# 入力チェックが必要か
	if {$buff2 != ""} {
	    destroy .t2
	}
    }
    pack .t2.l0 .t2.l1 .t2.e0
}

# end of file

