#! /usr/bin/wish
#
# ten_四川.tcl : 四川省風足して１０になるカードを取るパズル
#
#             Copyright (C) 1999,2000,2014,2019 by Makoto Hiroi
#
#
# 外部変数
#   board()   : 盤面（データを格納）
#               上位 4 ビットで色、下位 4 ビットで数字を表す
#   number()  : テキスト ID を格納
#   numstr()  : 表示する文字列
#   color()   : 表示する色
#   piece()   : 図形 ID を格納
#   px,py     : 選択した数字の座標( -1 は未選択 )
#   rest      : 残りの枚数
#   play_flag : ゲームの状態
#               0 : not play
#               1 : play 
#               2 : use search or takeback
#               4 : 検索中
#   buff1     : メッセージ表示用バッファ
#   buff2     : 名前入力用バッファ
#   time      : ゲーム開始時刻
#   id        : after コマンドが返す固有番号
#   history() : 取った牌の履歴
#   name()    : トップテン(1 - 10, 0 is dummy)
#   date()
#   score()
#

# サブルーチンリード
source "ten_sub.tcl"

# 盤面から色を求める
proc get_color {x y} {
    global board
    return [expr $board($x,$y) / 16]
}

# 盤面から数字を求める
proc get_number {x y} {
    global board
    return [expr $board($x,$y) % 16]
}

# ********** 判定ルーチン **********

# 四方向のチェック
proc check_close_piece {x y} {
    global board
    if {$board([expr $x + 1],$y) != 0 && $board([expr $x - 1],$y) != 0 &&
        $board($x,[expr $y + 1]) != 0 && $board($x,[expr $y - 1]) != 0} {
        return 1
    }
    return 0
}


# 隣り合ったカードか
proc check_neighbor {x1 y1 x2 y2} {
    set x [expr $x1 - $x2]
    set y [expr $y1 - $y2]
    if {($x == 0 && ($y == 1 || $y == -1)) ||
        ($y == 0 && ($x == 1 || $x == -1))} {
        return 1
    }
    return 0
}

# x 方向のチェック
proc check_xpath {x1 x2 y} {
    global board
    if {$x1 < $x2} {
	set x [expr $x1 + 1]
	set i $x2
    } else {
	set x [expr $x2 + 1]
	set i $x1
    }
    while {$x < $i} {
	if {$board($x,$y) != 0} {
	    return 0
	}
	incr x
    }
    return 1
}

# y 方向のチェック
proc check_ypath {y1 y2 x} {
    global board
    if {$y1 < $y2} {
	set y [expr $y1 + 1]
	set i $y2
    } else {
	set y [expr $y2 + 1]
	set i $y1
    }
    while {$y < $i} {
	if {$board($x,$y) != 0} {
	    return 0
	}
	incr y
    }
    return 1
}

# 二本の線でつながるか
proc check_two_line {x1 y1 x2 y2} {
    global board
    if {$board($x1,$y2) == 0} {
	if {[check_xpath $x1 $x2 $y2] && [check_ypath $y1 $y2 $x1]} {
	    return [concat $x1 $y1 $x1 $y2 $x2 $y2]
	}
    }
    if {$board($x2,$y1) == 0} {
	if {[check_xpath $x1 $x2 $y1] && [check_ypath $y1 $y2 $x2]} {
	    return [concat $x1 $y1 $x2 $y1 $x2 $y2]
	}
    }
    return ""
}

# 三本の線でつながるか
proc check_three_line_x {x1 y1 x2 y2 y} {
    global board
    if {$board($x1,$y) == 0 && $board($x2,$y) == 0} {
	if {[check_xpath $x1 $x2 $y] &&
	    [check_ypath $y1 $y $x1] &&
	    [check_ypath $y2 $y $x2] } {
	    return 1
	}
    }
    return 0
}

proc check_three_line_y {x1 y1 x2 y2 x} {
    global board
    if {$board($x,$y1) == 0 && $board($x,$y2) == 0} {
	if {[check_ypath $y1 $y2 $x] &&
	    [check_xpath $x1 $x $y1] &&
	    [check_xpath $x2 $x $y2]} {
	    return 1
	}
    }
    return 0
}

proc check_three_line {x1 y1 x2 y2} {
    # x 方向をサーチ
    if {$y1 < $y2} {
	set yl $y1
	set yh $y2
    } else {
	set yl $y2
	set yh $y1
    }
    for {set y $yl} {$y < $yh} {incr y} {
	if [check_three_line_x $x1 $y1 $x2 $y2 $y] {
            return [concat $x1 $y1 $x1 $y $x2 $y $x2 $y2]
	}
    }
    for {set y $yl} {$y >= 0} {incr y -1} {
	if [check_three_line_x $x1 $y1 $x2 $y2 $y] {
            return [concat $x1 $y1 $x1 $y $x2 $y $x2 $y2]
	}
    }
    for {set y $yh} {$y < 10} {incr y 1} {
	if [check_three_line_x $x1 $y1 $x2 $y2 $y] {
            return [concat $x1 $y1 $x1 $y $x2 $y $x2 $y2]
	}
    }
    # y 方向をサーチ
    if {$x1 < $x2} {
	set xl $x1
	set xh $x2
    } else {
	set xl $x2
	set xh $x1
    }
    for {set x $xl} {$x < $xh} {incr x} {
	if [check_three_line_y $x1 $y1 $x2 $y2 $x] {
            return [concat $x1 $y1 $x $y1 $x $y2 $x2 $y2]
	}
    }
    for {set x $xl} {$x >= 0} {incr x -1} {
	if [check_three_line_y $x1 $y1 $x2 $y2 $x] {
            return [concat $x1 $y1 $x $y1 $x $y2 $x2 $y2]
	}
    }
    for {set x $xh} {$x < 20} {incr x 1} {
	if [check_three_line_y $x1 $y1 $x2 $y2 $x] {
            return [concat $x1 $y1 $x $y1 $x $y2 $x2 $y2]
	}
    }
    return ""
}

# 削除可能か
proc check_remove_piece {x1 y1 x2 y2} {
    # 隣り合わせか
    if [check_neighbor $x1 $y1 $x2 $y2] {
	return [concat $x1 $y1 $x2 $y2]
    }
    # ４方向ともカードがあるか
    if {[check_close_piece $x1 $y1] || [check_close_piece $x2 $y2]} {
	return ""
    }
    # 一つの線でつながるか
    if {(($x1 == $x2) && [check_ypath $y1 $y2 $x1]) ||
        (($y1 == $y2) && [check_xpath $x1 $x2 $y1])} {
	return [concat $x1 $y1 $x2 $y2]
    }
    # 二本の線でつながるか
    set line [check_two_line $x1 $y1 $x2 $y2]
    if {$line == ""} {
	# 三本の線でつながるか
	set line [check_three_line $x1 $y1 $x2 $y2]
    }
    return $line
}

# 取れる牌を探す
proc search_piece {} {
    global board piece piece_postion piece_count color
    set result ""
    for {set c 0} {$c < 4} {incr c} {
	for {set n 1} {$n < 5} {incr n} {
	    set p1 [expr $c * 16 + $n]
	    set p2 [expr $c * 16 + (10 - $n)]
	    if {$piece_count($p1) > 0 && $piece_count($p2) > 0} {
		set i 0
		while {$i < 8} {
		    set x1 [lindex $piece_postion($p1) $i]
		    incr i
		    set y1 [lindex $piece_postion($p1) $i]
		    incr i
		    if {$board($x1,$y1) == 0} continue
		    set j 0
		    while {$j < 8} {
			set x2 [lindex $piece_postion($p2) $j]
			incr j
			set y2 [lindex $piece_postion($p2) $j]
			incr j
			if {$board($x2,$y2) == 0} continue
			if {[check_remove_piece $x1 $y1 $x2 $y2] != ""} {
			    set result [concat $result $piece($x1,$y1) $piece($x2,$y2)]
			}
		    }
		}
	    }
	}
	# 5 だけ特別
	set p1 [expr $c * 16 + 5]
	if {$piece_count($p1) > 0} {
	    set i 0
	    while {$i < 8} {
		set x1 [lindex $piece_postion($p1) $i]
		incr i
		set y1 [lindex $piece_postion($p1) $i]
		incr i
		if {$board($x1,$y1) == 0} continue
		set j $i
		while {$j < 8} {
		    set x2 [lindex $piece_postion($p1) $j]
		    incr j
		    set y2 [lindex $piece_postion($p1) $j]
		    incr j
		    if {$board($x2,$y2) == 0} continue
		    if {[check_remove_piece $x1 $y1 $x2 $y2] != ""} {
			set result [concat $result $piece($x1,$y1) $piece($x2,$y2)]
		    }
		}
	    }
	}
    }
    return $result
}

# 終了チェック
proc check_finish {} {
    global board rest time id buff2 play_flag
    if {$rest == 0} {
	# 終了
	set t [clock seconds]
	set s [expr $t - $time]
	after cancel $id
	set ranking [check_hi_score $s]
	if {$ranking > 0 && $play_flag == 1} {
	    input_hi_score_name $ranking 
	    tkwait window .t2
	    update_score $buff2 $t $s $ranking
	    write_score_file
	    open_score_window $ranking
	} else {
	    tk_messageBox -type ok \
		-message [format "おめでとう %s です" [change_seconds $s]]
	}
	return 1
    } else {
	# 手詰まりチェック
	if {[search_piece] == ""} {
	    after cancel $id
	    tk_messageBox -type ok -message "手詰まりです"
	    return 1
	}
    }
    return 0
}

# 線を引く
proc draw_line {point_list} {
    global back_id point_x point_y buffer
    set len [llength $point_list]
    set buffer "長さ $len"
    set cmd ".c0 create line"
    set i 0
    while {$i < $len} {
	set x [lindex $point_list $i]
	incr i
	set y [lindex $point_list $i]
	incr i
	set cmd [concat $cmd $point_x($x,$y) $point_y($x,$y)]
    }
    set cmd [concat $cmd "-fill white" "-width 10"]
    set id [eval $cmd]
    .c0 raise $id $back_id
    return $id
}

# 牌を取り除く
proc remove_piece {x y} {
    global board number rest piece piece_count history
    .c0 lower $number($x,$y)
    .c0 lower $piece($x,$y)
    .c0 itemconfigure $piece($x,$y) -fill white
    set p $board($x,$y)
    incr piece_count($p) -1
    set history($rest) [concat $p $x $y]
    set board($x,$y) 0
    incr rest -1
}


# 牌を戻す
proc takeback {} {
    global play_flag number piece board rest history px py
    if {$play_flag != 0 && $px != -1} {
	# 牌を選択している時はキャンセル扱いにする
	push_piece $px $py
	return
    }
    if {!($play_flag & 0x03) || $rest == 144} return
    set play_flag 2
    for {set i 0} {$i < 2} {incr i} {
	incr rest
	set h $history($rest)
	set p [lindex $h 0]
	set x [lindex $h 1]
	set y [lindex $h 2]
	.c0 raise $piece($x,$y)
	.c0 raise $number($x,$y)
	set board($x,$y) $p
    }
}

# 検索
proc search {} {
    global play_flag px py
    if {!($play_flag & 0x03)} return
    if {$px != -1} {
	# 選択していたらキャンセルして表示する
	push_piece $px $py
    }
    set play_flag 4
    set pieces [search_piece]
    set len [llength $pieces]
    set i 0
    while {$i < $len} {
	set p1 [lindex $pieces $i]
	incr i
	set p2 [lindex $pieces $i]
	incr i
	.c0 itemconfigur $p1 -fill darkgray
	.c0 itemconfigur $p2 -fill darkgray
	update
	after 1000
	.c0 itemconfigur $p1 -fill white
	.c0 itemconfigur $p2 -fill white
    }
    set play_flag 2
}

# バインド
proc push_piece {x y} {
    global board piece number play_flag px py
    if {!($play_flag & 0x03)} return
    if {$px == -1} {
	# 最初の選択
	set px $x
	set py $y
	.c0 itemconfigure $piece($x,$y) -fill darkgray
    } elseif {$px == $x && $py == $y} {
	# 同じであればキャンセル扱いにする
	.c0 itemconfigure $piece($x,$y) -fill white
	set px -1
	set py -1
    } else {
	set n1 [get_number $px $py]
	set c1 [get_color $px $py]
	set n2 [get_number $x $y]
	set c2 [get_color $x $y]
	if {$c1 == $c2 && [expr $n1 + $n2] == 10} {
	    set line [check_remove_piece $px $py $x $y]
	    if {$line != ""} {
		# 線を引いて削除する
		set save_flag $play_flag
                # 消去中の入力を禁止
		set play_flag 0
		.c0 itemconfigure $piece($x,$y) -fill darkgray
		set line_id [draw_line $line]
		update
		after 200
		# 削除する
		.c0 delete $line_id
		remove_piece $x $y
		remove_piece $px $py
		set px -1
		set py -1
                # 入力受け付け
		set play_flag $save_flag
		display_message
	    }
	}
	# 手詰まりのチェック
	if [check_finish] {
	    set play_flag 0
	}
    }
}



# ********** 初期化ルーチン **********

# 盤面の初期化
proc make_board {} {
    global board piece_postion piece_count
    # piece_table は局所変数
    for {set i 0; set c 0} {$c < 4} {incr c} {
	for {set n 1} {$n <= 9} {incr n} {
	    # piece_postion の初期化
	    set p [expr $c * 16 + $n]
	    set piece_postion($p) ""
	    set piece_count($p) 4
	    # piece_tabale の初期化
	    for {set j 0} {$j < 4} {incr j} {
		set piece_table($i) $p
		incr i
	    }
	}
    }
    # 乱数でかき回す
    for {set j 0} {$j < $i} {incr j} {
	set n    [expr int( rand() * $i )]
	set temp $piece_table($n)
	set piece_table($n) $piece_table($j)
	set piece_table($j) $temp
    }
    # board にセット
    set i 0
    for {set y 1} {$y <= 8} {incr y} {
	for {set x 1} {$x <= 18} {incr x} {
	    set p $piece_table($i)
	    set board($x,$y) $p
	    set piece_postion($p) [concat $piece_postion($p) $x $y]
	    incr i
	}
    }
}

# 盤面の描画
proc draw_board {} {
    global board number numstr piece color
    for {set y 1} {$y <= 8} {incr y} {
	for {set x 1} {$x <= 18} {incr x} {
	    set n [get_number $x $y]
	    set c [get_color $x $y]
	    .c0 itemconfigure $number($x,$y) \
		    -text $numstr($n) \
		    -fill $color($c)
	    .c0 raise $piece($x,$y)
	    .c0 raise $number($x,$y)
	}
    }
}

# メッセージの表示
proc display_message {} {
    global time buff1 rest
    set t [expr [clock seconds] - $time]
    set buff1 [format "残り %3d 枚    時間 %5s" $rest [change_seconds $t]]
}

# 時間の表示
proc display_time {} {
    global id
    display_message
    set id [after 1000 display_time]
}

# ゲームの開始
proc start_game {} {
    global play_flag time rest id px py pz
    if {$play_flag > 0} {
	after cancel $id
    }
    while 1 {
	make_board
	if {[search_piece] != ""} break
    }
    draw_board
    set rest 144
    set play_flag 1
    set time [clock seconds]
    set px -1
    set py -1
    display_time
}

# ********** メニューの設定 **********
menu .m -type menubar
. configure -menu .m
.m add cascade -label "Games"    -under 0 -menu .m.m1
.m add command -label "Takeback" -under 0 -command "takeback"
.m add command -label "Search"   -under 0 -command "search"
.m add command -label "Help"     -under 0 -command "help"
menu .m.m1 -tearoff no
.m.m1 add command -label "Start"   -under 0 -command "start_game"
.m.m1 add command -label "HiScore" -under 0 -command "open_score_window 0"
.m.m1 add separator
.m.m1 add command -label "Exit" -under 0 -command "exit"

# オプションの設定
option add *font "{Noto Sans Mono CJK JP} 12"

# 画面の設定
canvas .c0 -width 640 -height 320

# 背景
set back_id [.c0 create rectangle 0 0 639 319 -fill darkgreen]

# board, point の初期化
for {set y 0} {$y < 10} {incr y} {
    for {set x 0} {$x < 20} {incr x} {
	set board($x,$y) 0
	set point_x($x,$y) [expr $x * 32 + 16]
	set point_y($x,$y) [expr $y * 32 + 16]
    }
}

# カードの設定
for {set y 1} {$y <= 8} {incr y} {
    for {set x 1} {$x <= 18} {incr x} {
	set x1 [expr $x * 32]
	set x2 [expr $x1 + 31]
	set y1 [expr $y * 32]
	set y2 [expr $y1 + 31]
	set piece($x,$y) [.c0 create rectangle $x1 $y1 $x2 $y2 -fill white]
	set number($x,$y) [.c0 create text $point_x($x,$y) $point_y($x,$y) \
		 -text $numstr(0) \
			       -font "{Noto Sans Mono CJK JP} 24 bold"]

	.c0 bind $number($x,$y) <Button-1> "push_piece $x $y"
	.c0 bind $piece($x,$y) <Button-1> "push_piece $x $y"
    }
}

# 表示用ラベル
label .l1 -textvariable buff1 -bg darkgreen -fg white -anchor e

pack .l1 .c0 -fill x

# 窓の題名
wm title . "TEN"
wm resizable . 0 0

# 初期化
expr srand( [clock seconds] )
set play_flag 0
set path_name [file dirname $argv0]
set score_file "$path_name/TEN_四川.SCO"

# スコアファイルのリード
read_score_file
focus -force .

# end of file
